__version__ = "1.3.1"

# function get_bins()
# set dtype=int to fix DeprecationWarning __array__

# updated quartz weighing function

# fixed rounding error for low erosion rates:
#    get_textline(), get_textblock()
# remove leading and trailing spaces in data from spreadsheet file
#    .add_samples_from_dataframe(), .add_samples_from_file(), import_data()
# plotting functions returning fig, ax

# print_defaults(), print_standardizations()
# restandardize()

# Point-based erosion rates and exposure ages
# import_data(), get_textblock()
# get_ages_from_server(), get_erates_from_server()


